from .lib import SyncFHIRClient, AsyncFHIRClient

__title__ = "fhir-py"
__version__ = "1.4.1"
__author__ = "beda.software"
__license__ = "None"
__copyright__ = "Copyright 2023 beda.software"

# Version synonym
VERSION = __version__
